#!/usr/bin/env python3

#________________________________________________________________________________
# POLYNOMEN VERWERKEN MET NUMPY
#________________________________________________________________________________
__date__ = 20240420
__author__ = "Falk Mielke"

#________________________________________________________________________________
# BIBLIOTHEKEN
#________________________________________________________________________________
import numpy as NP
import matplotlib.pyplot as PLT


#________________________________________________________________________________
# DE POLYNOOM
#________________________________________________________________________________
class Polynoom(object):
    # een klas met alle nodige functies voor de bewerking van polynomen
    # zie verder: https://nl.wikipedia.org/wiki/Polynoom

    def __init__(self, coeffs = []):
        # de constructor van een polynoom met de vorm:
        #    y = f(x) = a₀ + a₁·x + a₂·x² + … + aᵢ·xⁱ
        # Op basis van coëfficiënten gaan we later alles kunnen analyseren.
        # INPUT:
        #     - coeffs: een lijst met coëfficiënten, van de laagste naar de hoogste exponent
        #

        # opslaan van de coëfficiënten
        self._c = coeffs
        # noot: het laag streepje (_) in het begin van een variable duidt aan die variable "privé" is,
        #       dus wij wensen niet dat die coëfficiënten van buiten de klas "Polynoom" aangepast worden.
        #       Denk aan de begrip "private parts" in het Engels.

        # de graad is de hoogste voorkomende macht van de variabele x,
        # dus hier eens minder dan de lengte van de coëfficiëntenreeks
        self._graad = len(self._c)-1


    def __str__(self):
        # Deze functie geeft een tekstbeschrijving van het concrete object terug,
        # bv. als we het willen afprinten.
        # zie ook: https://www.scaler.com/topics/python-str
        termen = [] # we gaan de termen apart versamelen
        for nr, c in enumerate(self._c):
            if c == 0:
                # lege coëfficiënt
                continue

            if nr == 0:
                # de constante
                termen.append(f"{c}")
                continue

            # voeg een term toe
            termen.append(("" if c == 1 else ("-" if c == -1 else  f"{c}·")) + f"x^{nr}")

        # hier gaan we de termen samenvoegen
        label = f"Polynoom {self._graad}e graad: y = " + " + ".join(termen)

        # negatieve coëfficiënten hebben geen "+" nodig
        label = label.replace(' + -', ' - ')

        # __str__ geeft altijd een string terug.
        return label


    def f(self, x):
        # hier krijgen we y-waardes voor (een reeks van) x-waardes.
        # we gebruiken de "polyval"-functie van numpy, MAAR met omgekeerde coëfficiënten-logica
        return NP.polyval(self._c[::-1], x)

    def afgeleide(self):
        # bereken de eerste afgeleide g(x) = df(x)/dx

        nieuwe_coeffs = []
        for nr, c in enumerate(self._c):
            # exponent vermindert om één
            if nr == 0:
                continue
            # ... en vermenigvuldigt met de coefficient
            nieuwe_coeffs.append(nr*c)

        return Polynoom(nieuwe_coeffs)


    def nulpunten(self, x):
        # deze functie bepaalt de nulpunten van het polynoom op een reeks x
        # door lineaire interpolatie van empirische nuldoorgangen
        # bron: https://stackoverflow.com/a/46911822
        # Noot: dit werkt niet voor contact- en zadelpunten
        # meer info:
        #  - https://math.libretexts.org/@go/page/15071
        #  - https://en.wikipedia.org/wiki/Polynomial_root-finding_algorithms
        #  - https://en.wikipedia.org/wiki/Aberth_method
        #  - https://en.wikipedia.org/wiki/Durand%E2%80%93Kerner_method
        y = self.f(x)
        s = NP.abs(NP.diff(NP.sign(y))).astype(bool)
        nulpunten = list(NP.unique(x[:-1][s] + NP.diff(x)[s]/(NP.abs(y[1:][s]/y[:-1][s])+1)))

        # kleine aanpassing om contactpunten te vinden
        afgeleide = self.afgeleide()
        y = afgeleide.f(x)
        s = NP.abs(NP.diff(NP.sign(y))).astype(bool)
        keerpunten = NP.unique(x[:-1][s] + NP.diff(x)[s]/(NP.abs(y[1:][s]/y[:-1][s])+1))
        for kp in keerpunten:
            if abs(self.f(kp)) < 1e-7:
                nulpunten.append(kp)

        return nulpunten


    def plot(self, x = None, ax = None, **plot_kwargs):
        # visualisatie van de polynoom,
        # indien er een bestaande canvas binnenkomt, gaan we die gebruiken.
        # De "plot_kwargs" geven eigenschappen zoals de kleur door aan de plot functie

        retoure = {} # hier gaan we optionele teruggaven versamelen
        if ax is None:
            cm = 1/2.54 # inch
            fig = PLT.figure(figsize = (16*cm, 12*cm), dpi = 300)
            ax = fig.add_subplot(1,1,1)
            retoure['fig'] = fig

        if x is None:
            x = NP.linspace(-1., 1., 201, endpoint = True)
            retoure['x'] = x

        ax.plot(x, self.f(x), **plot_kwargs)

        retoure['ax'] = ax # ter zekerheid geven we ook de canvas terug

        return retoure

#________________________________________________________________________________
# TESTING
#________________________________________________________________________________
# Hier testen we onze code van boven.
# Dit heeft enkele voordelen:
#  - we definiëren test cases waar de code goed werkt
#  - we anticiperen speciaal-gevallen en foutieve gebruiker-invoer
#  - we zien naar aanpassingen of het nog werkt
#  - we documenteren hoe iemand anders de code zou moeten gebruikt

# een reeks van coëfficiënten waarmee de polynoom moet werken.
_testbibliotheek = [  \
                       [] # een nep-polynoom, of "nulpolynoom"
                     , [1] # constante
                     , [-1, 0, 1] # x²-1
                     , [0, 0, 1] # x²-1
                     , [-1, 0, 0, 1] # x³-1
                     , [0, 0, 0, -1] # x³
                     , [-1, 1, 0.5, 2] # een iets langere term
                     ]


def TestTextLabels():
    # hier testen we de __str__() functie van polynomen

    # hier gaan we testen:
    for i, test in enumerate(_testbibliotheek):
        print("#"*16)
        print(f"   LABEL TEST {i}")
        print(Polynoom(test))


def TestNulpunten():
    # hier testen we de __str__() functie van polynomen

    x = NP.linspace(-2., 2., 401, endpoint = True)
    for i, test in enumerate(_testbibliotheek):
        print("#"*16)
        print(f"   NULPUNT TEST {i}")
        print(Polynoom(test).nulpunten(x))


def TestVisualisatie():
    # hier testen we de plot()-functie van polynomen testen

    cm = 1/2.54 # inch
    fig = PLT.figure(figsize = (16*cm, 12*cm), dpi = 300)
    ax = fig.add_subplot(1,1,1)
    x = NP.linspace(-2., 2., 401, endpoint = True)

    print("#"*16)
    print(f"   PLOT TESTS")

    for i, test in enumerate(_testbibliotheek):
        p = Polynoom(test)
        p.plot(x = x, ax = ax, ls = '-', lw = '1', label = str(p))

        # we duiden ook de nulpunten aan
        npt = p.nulpunten(x)
        ax.scatter(npt, NP.zeros_like(npt), s = 5, marker = 'o', facecolor = None, edgecolor = 'orange', lw = 0.5)

    ax.legend(fontsize = 8)
    ax.set_xlim([-2, 2])
    PLT.show()


def TestAfgeleide():
    # hier testen we de afgeleide van polynomen

    for i, test in enumerate(_testbibliotheek):
        print("#"*16)
        print(f"   AFGELEIDE TEST {i}")
        p = Polynoom(test)
        print(p, '  -->  ', p.afgeleide())


#________________________________________________________________________________
# MISSION CONTORL
#________________________________________________________________________________
if __name__ == "__main__":
    # dit is de code die we laten uitvoeren indien deze bestand niet als bibliotheek geïmporteerd wordt.

    # we gaan testen!
    TestTextLabels()
    TestNulpunten()
    TestVisualisatie()
    TestAfgeleide()
